@@ -15,17 +15,17 @@ import pandas as pd
 
 
 def get_npyx_memory(dp):
-    dprm = Path(dp,'npyxMemory')
-    old_dprm =Path(dp,'routinesMemory')
-    if old_dprm.exists():
-        print("Backward compatibility - renaming routinesMemory as npyxMemory.")
+    dpnm = Path(dp) / 'npyxMemory'
+    old_dpnm = Path(dp) / 'routinesMemory'
+    if old_dpnm.exists() and not dpnm.exists():
         try: # because of parallel proccessing, might have been renamed in the process!
-            os.rename(str(old_dprm), str(dprm))
+            os.rename(str(old_dpnm), str(dpnm))
+            print("Backward compatibility - renaming routinesMemory as npyxMemory.")
         except:
-            assert dprm.exists()
-    if not os.path.isdir(dprm): os.makedirs(dprm)
+            assert dpnm.exists()
+    os.makedirs(dpnm, exist_ok=True)
 
-    return dprm
+    return dpnm
 
 def get_datasets(ds_master, ds_paths_master, ds_behav_master=None, warnings=True):
     """
@@ -180,6 +180,10 @@ def detect_new_spikesorting(dp, print_message=True, qualities=None):
     '''
     dp=Path(dp)
     spikesorted = False
+    if not (dp/'cluster_group.tsv').exists():
+        # if first time this is ran on a merged dataset
+        return True
+    
     if qualities is None:
         assert 'merged' not in str(dp), 'this function should be ran on an original sorted dataset, not on a merged dataset.'
         last_spikesort = os.path.getmtime(dp/'spike_clusters.npy')
@@ -315,6 +319,6 @@ def get_good_units(dp):
     return get_units(dp, quality='good')
 
 # circular imports
-from npyx.io import read_metadata
+from npyx.inout import read_metadata
 from npyx.spk_wvf import get_depthSort_peakChans
 from npyx.merger import assert_multi, get_ds_table
